/* 
	RFCOMMd - RFCOMM daemon.
	Copyright (C) 2001 Qualcomm Incorporated

	Written 2001 by Maxim Krasnyansky <maxk@qualcomm.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 2 as
	published by the Free Software Foundation;

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
	IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY CLAIM,
	OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER
	RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
	NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
	USE OR PERFORMANCE OF THIS SOFTWARE.

	ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS, COPYRIGHTS,
	TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS SOFTWARE IS DISCLAIMED.
*/

/* 
	Based on:
	VTun - Virtual Tunnel over TCP/IP network.
	Copyright (C) 1998-2001 Maxim Krasnyansky <max_mk@yahoo.com>
*/

/*
 * $Id$
 */

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <syslog.h>
#include <errno.h>

#include "rfcommd.h"
#include "lib.h"

#ifndef HAVE_SETPROC_TITLE
/* Functions to manipulate with program title */

extern char **environ;
char	*title_start;	/* start of the proc title space */
char	*title_end;     /* end of the proc title space */
int	title_size;

void init_title(int argc,char *argv[], char *envp[], char *name)
{
	int i;

	/*
	 *  Move the environment so settitle can use the space at
	 *  the top of memory.
	 */

	for (i = 0; envp[i]; i++);

	environ = (char **) malloc(sizeof (char *) * (i + 1));

	for(i = 0; envp[i]; i++)
		environ[i] = strdup(envp[i]);
	environ[i] = NULL;

	/*
	 *  Save start and extent of argv for set_title.
	 */

	title_start = argv[0];

	/*
	 *  Determine how much space we can use for set_title.  
	 *  Use all contiguous argv and envp pointers starting at argv[0]
		 */
	for(i=0; i<argc; i++)
		if( !i || title_end == argv[i])
			title_end = argv[i] + strlen(argv[i]) + 1;

	for(i=0; envp[i]; i++)
		if( title_end == envp[i] )
			title_end = envp[i] + strlen(envp[i]) + 1;

	strcpy(title_start, name);
	title_start += strlen(name);
	title_size = title_end - title_start;
}

void set_title(const char *fmt, ...)
{
	char buf[255];
	va_list ap;

	memset(title_start,0,title_size);

	/* print the argument string */
	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	if( strlen(buf) > title_size - 1)
		buf[title_size - 1] = '\0';

	strcat(title_start, buf);
}
#endif  /* HAVE_SETPROC_TITLE */

/*
 * Substitutes opt in place off '%X'. 
 * Returns new string.
 */
char * subst_opt(char *str, struct session *s)
{
	struct subst *opt = &s->sopt;
	register int slen, olen, sp, np;
	register char *optr, *nstr, *tmp;
	char buf[10];

	if( !str ) return NULL;

	slen = strlen(str) + 1;
	if( !(nstr = malloc(slen)) )
		return str;

	sp = np = 0;
	while( str[sp] ) {
		switch( str[sp] ) {
			case '%':
				optr = NULL;
				/* Check supported opt */
				switch( str[sp+1] ) {
					case '%':
					case 'd':
						optr=opt->dev;
						break;
					case 'N':
						optr=s->name;
						break;
					case 'A':
						optr=opt->laddr;
						break;
					case 'P':
						sprintf(buf,"%d",opt->lpsm);
						optr=buf;
						break;
					case 'a':
						optr=opt->raddr;
						break;
					case 'p':
						sprintf(buf,"%d",opt->rpsm);
						optr=buf;
						break;
					default:
						sp++;
						continue;
				}
				if( optr ) {
					/* Opt found substitute */
					olen = strlen(optr);
					slen = slen - 2 + olen;
					if( !(tmp = realloc(nstr, slen)) ) {
						free(nstr);
						return str;
					}
					nstr = tmp;
					memcpy(nstr + np, optr, olen);
					np += olen;
				}
				sp += 2;
				continue;

			case '\\':
				nstr[np++] = str[sp++];
				if( !nstr[sp] )
					continue;
				/* fall through */
			default:
				nstr[np++] = str[sp++];
				break;
		}
	}
	nstr[np] = '\0';
	return nstr;
}

/*
 * Split arguments string.
 * ' ' - group arguments
 * Modifies original string. 
 */
void split_args(char *str, char **argv)
{
	register int i = 0;
	int mode = 0;

	while( str && *str ) {
		switch( *str ) {
			case ' ':
				if( mode == 1 ) {
					*str = '\0';
					mode = 0;
					i++;
				}
				break;

			case '\'':
				if( !mode ) {
					argv[i] = str+1;
					mode = 2;
				} else {
					memmove(argv[i]+1, argv[i], str - argv[i]);
					argv[i]++;

					if( mode == 1 )
						mode = 2;
					else
						mode = 1;
				}
				break;

			case '\\':
				if( mode ) {
					memmove(argv[i]+1, argv[i], str - argv[i]);
					argv[i]++;
				}
				if( !*(++str) ) continue;
				/*Fall through */

			default:
				if( !mode ) {
					argv[i] = str;
					mode = 1;
				}
				break;
		}
		str++;
	}
	if( mode == 1 || mode == 2)
		i++;

	argv[i]=NULL;
}

int run_cmd(void *c, void *d)
{
	struct command *cmd = c;
	char *argv[50], *args;
	int pid, st;

	switch ((pid=fork())) {
		case 0:
			break;
		case -1:
			syslog(LOG_ERR,"Couldn't fork()");
			return 0;
		default:
			if(cmd->flags & CMD_WAIT) {
			        DEBUG_LOG("run_cmd parent waits \"%s\"",
					cmd->args);
				/* Wait for termination */
				if( waitpid(pid,&st,0) > 0 && (WIFEXITED(st) && WEXITSTATUS(st)) )
					syslog(LOG_INFO,"Command [%s %.20s] error %d",
					       cmd->prog ? cmd->prog : "sh",
					       cmd->args ? cmd->args : "",
					       WEXITSTATUS(st) );
			}
			if (cmd->flags & CMD_DELAY) {
				struct timespec tm = { CMD_DELAY_SEC, 0 };
				nanosleep(&tm, NULL);
			}
			DEBUG_LOG("run_cmd parent returns on or after "
				" \"%s\"", cmd->args);
			return 0;
	}

	DEBUG_LOG("execv() in child");
	args = subst_opt(cmd->args, d);
	if (!cmd->prog) {
		/* Run using shell */
		cmd->prog = "/bin/sh";
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = args;
		argv[3] = NULL;
	} else {
		argv[0] = cmd->prog;
		split_args(args, argv + 1);
	}
	execv(cmd->prog, argv);

	syslog(LOG_ERR,"Couldn't exec program %s", cmd->prog);
	exit(1);
}

void free_sopt(struct subst *opt)
{
	if (opt->dev) {
		free(opt->dev);
		opt->dev = NULL;
	}

	if (opt->laddr) {
		free(opt->laddr);
		opt->laddr = NULL;
	}

	if (opt->raddr) {
		free(opt->raddr);
		opt->raddr = NULL;
	}
};
