/* 
	RFCOMMd - RFCOMM daemon.
	Copyright (C) 2001 Qualcomm Incorporated

	Written 2001 by Maxim Krasnyansky <maxk@qualcomm.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 2 as
	published by the Free Software Foundation;

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
	IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY CLAIM,
	OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER
	RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
	NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
	USE OR PERFORMANCE OF THIS SOFTWARE.

	ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS, COPYRIGHTS,
	TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS SOFTWARE IS DISCLAIMED.
*/

/* 
	Based on:
	VTun - Virtual Tunnel over TCP/IP network.
	Copyright (C) 1998-2001 Maxim Krasnyansky <max_mk@yahoo.com>
*/

/*
 * $Id$
 */

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <strings.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <syslog.h>
#include <time.h>

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_SCHED_H
#include <sched.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#include "rfcommd.h"
#include "linkfd.h"
#include "lib.h"
#include "driver.h"

/* Host we are working with.
 * Used by signal handlers that's why it is global. 
 */
struct rfcommd_host *lfd_host;

struct lfd_mod *lfd_mod_head = NULL, *lfd_mod_tail = NULL;

/* Modules functions*/

/* Add module to the end of modules list */
void lfd_add_mod(struct lfd_mod *mod)
{
	if( !lfd_mod_head )
	{
		lfd_mod_head = lfd_mod_tail = mod;
		mod->next = mod->prev = NULL;
	} else
	{
		lfd_mod_tail->next = mod;
		mod->prev = lfd_mod_tail;
		mod->next = NULL;
		lfd_mod_tail = mod;
	}
}

/*  Initialize and allocate each module */
int lfd_alloc_mod(struct rfcommd_host *host)
{
	struct lfd_mod *mod = lfd_mod_head;

	while( mod )
	{
		if( mod->alloc && (mod->alloc)(host) )
			return 1;
		mod = mod->next;
	}

	return 0;
}

/* Free all modules */
int lfd_free_mod(void)
{
	struct lfd_mod *mod = lfd_mod_head;

	while( mod ) {
		if( mod->free && (mod->free)() )
			return 1;
		mod = mod->next;
	}
	lfd_mod_head = lfd_mod_tail = NULL;
	return 0;
}

/* Run modules down (from head to tail) */
inline int lfd_run_down(int len, char *in, char **out)
{
	register struct lfd_mod *mod;

	*out = in;
	for(mod = lfd_mod_head; mod && len > 0; mod = mod->next )
		if( mod->encode ) {
			len = (mod->encode)(len, in, out);
			in = *out;
		}
	return len;
}

/* Run modules up (from tail to head) */
inline int lfd_run_up(int len, char *in, char **out)
{
	register struct lfd_mod *mod;

	*out = in;
	for(mod = lfd_mod_tail; mod && len > 0; mod = mod->prev )
		if( mod->decode ) {
			len = (mod->decode)(len, in, out);
			in = *out;
		}
	return len;
}

/* Check if modules are accepting the data(down) */
inline int lfd_check_down(void)
{
	register struct lfd_mod *mod;
	int err = 1;

	for(mod = lfd_mod_head; mod && err > 0; mod = mod->next )
		if( mod->avail_encode )
			err = (mod->avail_encode)();
	return err;
}

/* Check if modules are accepting the data(up) */
inline int lfd_check_up(void)
{
	register struct lfd_mod *mod;
	int err = 1;

	for(mod = lfd_mod_tail; mod && err > 0; mod = mod->prev)
		if( mod->avail_decode )
			err = (mod->avail_decode)();

	return err;
}

/********** Linker *************/
/* Termination flag */
static int linker_term;

static void sig_term(int sig)
{
	syslog(LOG_INFO,"Closing connection");
	linker_term = 1;
}

/* Statistic dump */
void sig_alarm(int sig)
{
	static time_t tm;
	static char stm[20];

	tm = time(NULL);
	strftime(stm, sizeof(stm)-1, "%b %d %H:%M:%S", localtime(&tm));
	fprintf(lfd_host->stat.file,"%s %lu %lu %lu %lu\n", stm,
	        lfd_host->stat.byte_in, lfd_host->stat.byte_out,
	        lfd_host->stat.comp_in, lfd_host->stat.comp_out);

	alarm(RFCOMMD_STAT_IVAL);
}

static void sig_usr1(int sig)
{
	/* Reset statistic counters on SIGUSR1 */
	lfd_host->stat.byte_in = lfd_host->stat.byte_out = 0;
	lfd_host->stat.comp_in = lfd_host->stat.comp_out = 0;
}

int lfd_linker(void)
{
	int fd1 = lfd_host->rmt_fd;
	int fd2 = lfd_host->loc_fd;
	register int len, fl;
	struct timeval tv;
	char *buf, *out;
	fd_set fdset;
	int maxfd, idle = 0;

	if( !(buf = lfd_alloc(RFCOMMD_FRAME_SIZE + RFCOMMD_FRAME_OVERHEAD)) ) {
		syslog(LOG_ERR,"Can't allocate buffer for the linker");
		return 0;
	}

	proto_write(fd1, buf, RFCOMMD_ECHO_REQ);

	maxfd = (fd1 > fd2 ? fd1 : fd2) + 1;

	linker_term = 0;
	while( !linker_term ) {
		errno = 0;

		/* Wait for data */
		FD_ZERO(&fdset);
		FD_SET(fd1, &fdset);
		FD_SET(fd2, &fdset);

		tv.tv_sec  = lfd_host->ka_interval;
		tv.tv_usec = 0;

		if( (len = select(maxfd, &fdset, NULL, NULL, &tv)) < 0 ) {
			if( errno != EAGAIN && errno != EINTR )
				break;
			else
				continue;
		}

		if( !len ) {
			/* We are idle, lets check connection */
			if( lfd_host->flags & RFCOMMD_KEEP_ALIVE ) {
				if( ++idle > lfd_host->ka_failure ) {
					syslog(LOG_INFO,"Session %s network timeout", lfd_host->host);
					break;
				}
				/* Send ECHO request */
				if( proto_write(fd1, buf, RFCOMMD_ECHO_REQ) < 0 )
					break;
			}
			continue;
		}

		/* Read frames from network(fd1), decode and pass them to
		        * the local device (fd2) */
		if( FD_ISSET(fd1, &fdset) && lfd_check_up() ) {
			idle = 0;
			if( (len=proto_read(fd1, buf)) <= 0 )
				break;

			/* Handle frame flags */
			fl = len & ~RFCOMMD_FSIZE_MASK;
			len = len & RFCOMMD_FSIZE_MASK;
			if( fl ) {
				if( fl==RFCOMMD_BAD_FRAME ) {
					syslog(LOG_ERR, "Received bad frame");
					continue;
				}
				if( fl==RFCOMMD_ECHO_REQ ) {
					/* Send ECHO reply */
					if( proto_write(fd1, buf, RFCOMMD_ECHO_REP) < 0 )
						break;
					continue;
				}
				if( fl==RFCOMMD_ECHO_REP ) {
					/* Just ignore ECHO reply */
					continue;
				}
				if( fl==RFCOMMD_CONN_CLOSE ) {
					syslog(LOG_INFO,"Connection closed by other side");
					break;
				}
			}

			lfd_host->stat.comp_in += len;
			if( (len=lfd_run_up(len,buf,&out)) == -1 )
				break;
			if( len && dev_write(fd2,out,len) < 0 ) {
				if( errno != EAGAIN && errno != EINTR )
					break;
				else
					continue;
			}
			lfd_host->stat.byte_in += len;
		}

		/* Read data from the local device(fd2), encode and pass it to
		        * the network (fd1) */
		if( FD_ISSET(fd2, &fdset) && lfd_check_down() ) {
			if( (len = dev_read(fd2, buf, RFCOMMD_FRAME_SIZE)) < 0 ) {
				if( errno != EAGAIN && errno != EINTR )
					break;
				else
					continue;
			}
			if( !len ) break;

			lfd_host->stat.byte_out += len;
			if( (len=lfd_run_down(len,buf,&out)) == -1 )
				break;
			if( len && proto_write(fd1, out, len) < 0 )
				break;
			lfd_host->stat.comp_out += len;
		}
	}
	if( !linker_term && errno )
		syslog(LOG_INFO,"%s (%d)", strerror(errno), errno);

	/* Notify other end about our close */
	proto_write(fd1, buf, RFCOMMD_CONN_CLOSE);
	lfd_free(buf);

	return 0;
}

/* Link remote and local file descriptors */
int linkfd(struct rfcommd_host *host)
{
	struct sigaction sa, sat;
	int old_prio;

	lfd_host = host;

	old_prio=getpriority(PRIO_PROCESS,0);
	setpriority(PRIO_PROCESS,0,LINKFD_PRIO);

	/* Build modules stack */
	if(host->flags & RFCOMMD_ZLIB)
		lfd_add_mod(&lfd_zlib);

	if(host->flags & RFCOMMD_LZO)
		lfd_add_mod(&lfd_lzo);

	if(host->flags & RFCOMMD_ENCRYPT)
		lfd_add_mod(&lfd_encrypt);

	if(host->flags & RFCOMMD_SHAPE)
		lfd_add_mod(&lfd_shaper);

	if(lfd_alloc_mod(host))
		return 0;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler=sig_term;
	sigaction(SIGTERM,&sa,&sat);

	/* Initialize statstic dumps */
	if( host->flags & RFCOMMD_STAT )
	{
		char file[40];

		sa.sa_handler=sig_alarm;
		sigaction(SIGALRM,&sa,NULL);
		sa.sa_handler=sig_usr1;
		sigaction(SIGUSR1,&sa,NULL);

		sprintf(file,"%s/%.20s", RFCOMMD_STAT_DIR, host->host);
		if( (host->stat.file=fopen(file, "a")) ) {
			setvbuf(host->stat.file, NULL, _IOLBF, 0);
			alarm(RFCOMMD_STAT_IVAL);
		} else
			syslog(LOG_ERR, "Can't open stats file %s", file);
	}

	lfd_linker();

	if( host->flags & RFCOMMD_STAT )
	{
		alarm(0);
		fclose(host->stat.file);
	}

	lfd_free_mod();

	sigaction(SIGTERM,&sat,NULL);

	setpriority(PRIO_PROCESS,0,old_prio);

	return linker_term;
}
