/* 
	RFCOMMd - RFCOMM daemon.
	Copyright (C) 2001 Qualcomm Incorporated

	Written 2001 by Maxim Krasnyansky <maxk@qualcomm.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 2 as
	published by the Free Software Foundation;

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
	IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY CLAIM,
	OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER
	RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
	NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
	USE OR PERFORMANCE OF THIS SOFTWARE.

	ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS, COPYRIGHTS,
	TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS SOFTWARE IS DISCLAIMED.
*/

/* 
	Based on:
	VTun - Virtual Tunnel over TCP/IP network.
	Copyright (C) 1998-2001 Maxim Krasnyansky <max_mk@yahoo.com>
*/

/*
 * $Id$
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>

#include "rfcomm.h"
#include "rfcommd.h"
#include "lib.h"

/* Global options for the server and client */
struct options rfcommd;
struct session default_session;

volatile sig_atomic_t __io_canceled;

void write_pid(void);
void reread_config(int sig);
void usage(void);

extern int optind,opterr,optopt;
extern char *optarg;

int main(int argc, char *argv[], char *env[])
{
	int svr, daemon, dofork, opt;
	struct session *session = NULL;
	struct sigaction sa;

	/* Ignore deadly signals */
	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = SIG_IGN;
	sa.sa_flags = SA_NOCLDWAIT;
	sigaction(SIGQUIT, &sa, NULL);
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);
	sigaction(SIGHUP, &sa, NULL);
	sigaction(SIGUSR1, &sa, NULL);

	/* Configure default settings */
	svr = 0; daemon = 1; dofork = 1;

	DEBUG_MARK();

	rfcommd.cfg_file = RFCOMMD_CONFIG_FILE;
	rfcommd.timeout = -1;

	/* Dup strings because parser will try to free them */
	rfcommd.ppp   = strdup("/usr/sbin/pppd");
	rfcommd.ifcfg = strdup("/sbin/ifconfig");
	rfcommd.route = strdup("/sbin/route");
	rfcommd.fwall = strdup("/usr/sbin/iptables");

	rfcommd.svr_addr = NULL;
	rfcommd.svr_psm  = -1;

	/* Initialize default session options */
	memset(&default_session, 0, sizeof(default_session));
	default_session.multi   = RFCOMMD_MULTI_ALLOW;
	default_session.timeout = RFCOMMD_CONNECT_TIMEOUT;

	/* Start logging to syslog and stderr */
	openlog("rfcommd", LOG_PID | LOG_NDELAY | LOG_PERROR, LOG_DAEMON);

	DEBUG_MARK();

	while( (opt=getopt(argc,argv,"isf:P:t:np")) != EOF ) {
		switch(opt) {
			case 's':
				svr = 1;
				break;
			case 'P':
				rfcommd.svr_psm  = atoi(optarg);
				break;
			case 'f':
				rfcommd.cfg_file = strdup(optarg);
				break;
			case 'n':
				daemon = 0;
				break;
			case 't':
				rfcommd.timeout = atoi(optarg);
				break;
			default:
				usage();
				exit(1);
		}
	}

	DEBUG_MARK();

	reread_config(0);

	DEBUG_MARK();

	if( !svr ){
		char *ses;
		if( argc - optind < 2 ) {
			usage();
			exit(1);
		}
		ses = argv[optind++];

		if( !(session = get_session_by_name(ses)) ){
			syslog(LOG_ERR,"Session %s not found in %s", ses, rfcommd.cfg_file);
			exit(1);
		}

		rfcommd.svr_addr = strdup(argv[optind]);
	}

	/*
	 * Now fill uninitialized fields of the options structure
	 * with default values. 
	 */
	if(rfcommd.svr_psm == -1)
		rfcommd.svr_psm = RFCOMM_PSM;
	if(rfcommd.timeout == -1)
		rfcommd.timeout = RFCOMMD_TIMEOUT;

	if( daemon ){
		int fd;

		if( dofork && fork() )
			exit(0);

		/* Direct stdin,stdout,stderr to '/dev/null' */
		close(0); close(1); close(2);
		fd = open("/dev/null", O_RDWR);
		if (fd) {
			dup2(fd, 0);
			close(fd);
		}
		dup2(0, 1); dup2(0, 2);

		setsid();

		chdir("/");
	}

	DEBUG_MARK();

	io_init();

	DEBUG_MARK();

	if( svr ){
		sa.sa_handler = reread_config;
		sigaction(SIGHUP, &sa, NULL);

		init_title(argc, argv, env, "rfcommd[s]: ");
		write_pid();
		server();
	} else {
		init_title(argc, argv, env, "rfcommd[c]: ");
		client(session);
	}

	DEBUG_MARK();

	closelog();

	DEBUG_MARK();
	return 0;
}

/*
 * Very simple PID file creation function. Used by server.
 * Overrides existing file. 
 */
void write_pid(void)
{
	FILE *f;

	if( !(f=fopen(RFCOMMD_PID_FILE,"w")) ) {
		syslog(LOG_ERR,"Can't write PID file");
		return;
	}

	fprintf(f,"%d",(int)getpid());
	fclose(f);
}

void reread_config(int sig)
{
	if( !read_config(rfcommd.cfg_file) ) {
		syslog(LOG_ERR,"No sessions defined");
		exit(1);
	}
}

void usage(void)
{
	printf("RFCOMMd ver %s\n", RFCOMMD_VER);
	printf("Usage: \n");
	printf("  Server:\n");
	printf("\trfcommd <-s> [-n] [-r] [-f file] [-P psm]\n");
	printf("  Client:\n");
	printf("\trfcommd [-n] [-f file] [-P psm] [-t timeout] <session> <server bd adress>\n");
}
