%{
/* 
	RFCOMMd - RFCOMM daemon.
	Copyright (C) 2001 Qualcomm Incorporated

	Written 2001 by Maxim Krasnyansky <maxk@qualcomm.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 2 as
	published by the Free Software Foundation;

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
	IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY CLAIM,
	OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER
	RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
	NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
	USE OR PERFORMANCE OF THIS SOFTWARE.

	ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS, COPYRIGHTS,
	TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS SOFTWARE IS DISCLAIMED.
*/

/* 
	Based on:
	VTun - Virtual Tunnel over TCP/IP network.
	Copyright (C) 1998-2001 Maxim Krasnyansky <max_mk@yahoo.com>
*/

/*
 * $Id$
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <signal.h>

#include "lib.h"
#include "rfcommd.h"
#include "kword.h"

int lineno = 1;

struct session *parse_session;
extern struct session default_session;

llist  *parse_cmds;
struct command parse_cmd;

llist sessions;

int cfg_error(const char *fmt, ...);
int add_cmd(llist *cmds, char *prog, char *args, int flags);
void *cp_cmd(void *d, void *u);
int free_cmd(void *d, void *u);

void copy_addr(struct session *to, struct session *from);
void free_addr(struct session *s);

void free_sessions(void);

int yyparse(void);
int yylex(void);	
int yyerror(char *s); 

#define YYERROR_VERBOSE 1

%}

%union {
   char *str;
   int  num;
   struct { int num1; int num2; } dnum;
}

%token K_OPTIONS K_DEFAULT K_PSM K_TIMEOUT K_PERSIST K_CHANNEL
%token K_PROG K_PPP K_SPEED K_IFCFG K_FWALL K_ROUTE K_DEVICE 
%token K_MULTI K_SRCADDR K_IFACE K_ADDR
%token K_STAT K_UP K_DOWN

%token <str> K_SESSION K_ERROR
%token <str> WORD PATH STRING
%token <num> NUM 
%token <dnum> DNUM

%%
config: 
  | config statement 
  ;

statement: '\n'
  | K_OPTIONS   '{' options '}' 

  | K_DEFAULT   { 
		  parse_session = &default_session; 
                }        
    '{' session_options '}' 
  | K_SESSION      { 
		  if( !(parse_session = malloc(sizeof(struct session))) ){
		     yyerror("No memory for the session");
		     YYABORT;
		  }

		  /* Fill new session struct with default values.
		   * MUST dup strings to be able to reread config.
		   */
	  	  memcpy(parse_session, &default_session, sizeof(struct session));
		  parse_session->name = strdup($1);

		  /* Copy local address */
		  copy_addr(parse_session, &default_session);

		  llist_copy(&default_session.up, &parse_session->up, cp_cmd, NULL);
		  llist_copy(&default_session.down, &parse_session->down, cp_cmd, NULL);

		  /* Add session to the list */
		  llist_add(&sessions, (void *)parse_session);
		}    
    '{' session_options '}' 
  | K_ERROR	{
		  cfg_error("Invalid clause '%s'",$1);
		  YYABORT;
		}
  ;

options:
    option
  | options option
  ;

/* Don't override command line options */
option:  '\n'
  | K_PSM NUM 		{ 
			  if(rfcommd.svr_psm == -1)
			     rfcommd.svr_psm = $2;
			} 
  | K_TIMEOUT NUM 	{  
			  if(rfcommd.timeout == -1)
			     rfcommd.timeout = $2; 	
			}
  | K_PPP   PATH	{
			  free(rfcommd.ppp);
			  rfcommd.ppp = strdup($2);
			}
  | K_IFCFG PATH	{
			  free(rfcommd.ifcfg);
			  rfcommd.ifcfg = strdup($2);
			}
  | K_ROUTE PATH 	{   
			  free(rfcommd.route);  
			  rfcommd.route = strdup($2); 	
			}		
  | K_FWALL PATH 	{   
			  free(rfcommd.fwall);  
			  rfcommd.fwall = strdup($2); 	
			}
  | K_ERROR		{
			  cfg_error("Unknown option '%s'",$1);
			  YYABORT;
			}
  ;

session_options:
    session_option
  | session_options session_option
  ;

/* Host options. Must free strings first, because they 
 * could be strduped from default_session */
session_option: '\n'
  | K_MULTI NUM		{ 
			  parse_session->multi = $2;
			}
  | K_TIMEOUT NUM	{ 
			  parse_session->timeout = $2;
			}
  | K_CHANNEL NUM	{ 
			  parse_session->channel = $2;
			}
  | K_STAT NUM		{
			  if( $2 )
			     parse_session->flags |= RFCOMMD_STAT;
			  else
			     parse_session->flags &= ~RFCOMMD_STAT;
			}
  | K_PERSIST NUM 	{ 
			  if( $2 ) 
			     parse_session->flags |= RFCOMMD_PERSIST;
			  else	
			     parse_session->flags &= ~RFCOMMD_PERSIST;
			}
  | K_SRCADDR 		'{' srcaddr_options '}'

  | K_UP 	        { 
			  parse_cmds = &parse_session->up; 
   			  llist_free(parse_cmds, free_cmd, NULL);   
			} '{' command_options '}' 
  | K_DOWN 	        { 
			  parse_cmds = &parse_session->down; 
   			  llist_free(parse_cmds, free_cmd, NULL);   
			} '{' command_options '}' 
  | K_ERROR		{
			  cfg_error("Unknown option '%s'",$1);
			  YYABORT;
			} 
  ;


srcaddr_options: /* empty */
  | srcaddr_option
  | srcaddr_options srcaddr_option
  ;

srcaddr_option:  
  K_ADDR WORD		{
			  free_addr(parse_session);
			  parse_session->src_addr.name = strdup($2);
			  parse_session->src_addr.type = RFCOMMD_ADDR_NAME;
			}
  | K_IFACE WORD	{
			  free_addr(parse_session);
			  parse_session->src_addr.name = strdup($2);
			  parse_session->src_addr.type = RFCOMMD_ADDR_IFACE;
			}
  | K_IFACE STRING	{
			  free_addr(parse_session);
			  parse_session->src_addr.name = strdup($2);
			  parse_session->src_addr.type = RFCOMMD_ADDR_IFACE;
			}
  | K_ERROR		{
			  cfg_error("Unknown option '%s'",$1);
			  YYABORT;
			} 
  ;

command_options: /* empty */
  | command_option
  | command_options command_option
  ;

command_option: '\n' 
  | K_PROG		{
			  memset(&parse_cmd, 0, sizeof(struct command));
			} 
 	prog_options    {
			  add_cmd(parse_cmds, parse_cmd.prog, parse_cmd.args, parse_cmd.flags);
			}
  | K_PPP STRING 	{   
			  add_cmd(parse_cmds, strdup(rfcommd.ppp), strdup($2),
				0);
			}		
  | K_IFCFG STRING 	{   
			  add_cmd(parse_cmds, strdup(rfcommd.ifcfg),strdup($2), CMD_WAIT);
			}
  | K_ROUTE STRING 	{   
			  add_cmd(parse_cmds, strdup(rfcommd.route),strdup($2), CMD_WAIT);
			}
  | K_FWALL STRING 	{   
			  add_cmd(parse_cmds, strdup(rfcommd.fwall),strdup($2), CMD_WAIT);
			}
  | K_ERROR		{
			  cfg_error("Unknown cmd '%s'",$1);
			  YYABORT;
			} 
  ;

prog_options:
    prog_option
  | prog_options prog_option
  ;

prog_option:
  PATH  		{
			  parse_cmd.prog = strdup($1);
			}
  | STRING 		{
			  parse_cmd.args = strdup($1);
			}
  | NUM		   	{
			  parse_cmd.flags = $1;
			}
  ;
%%

int yyerror(char *s) 
{
   syslog(LOG_ERR, "%s line %d\n", s, lineno);
   return 0;
}

int cfg_error(const char *fmt, ...)
{
   char buf[255];
   va_list ap;

   /* print the argument string */
   va_start(ap, fmt);
   vsnprintf(buf,sizeof(buf),fmt,ap);
   va_end(ap);

   yyerror(buf);
   return 0;
}

int add_cmd(llist *cmds, char *prog, char *args, int flags)
{
   struct command *cmd;
   if( !(cmd = malloc(sizeof(struct command))) ){
      yyerror("No memory for the command");
      return -1;
   }
   memset(cmd, 0, sizeof(struct command)); 		   			
   cmd->prog = prog;
   cmd->args = args;
   cmd->flags = flags;
   llist_add(cmds, cmd);

   return 0;
}		

void *cp_cmd(void *d, void *u)
{
   struct command *cmd = d, *cmd_copy; 

   if( !(cmd_copy = malloc(sizeof(struct command))) ){
      yyerror("No memory to copy command");
      return NULL;
   }
 
   cmd_copy->prog = strdup(cmd->prog);
   cmd_copy->args = strdup(cmd->args);
   cmd_copy->flags = cmd->flags;
   return cmd_copy;
}

int free_cmd(void *d, void *u)
{
   struct command *cmd = d; 
   free(cmd->prog);
   free(cmd->args);
   free(cmd);
   return 0;
}

void copy_addr(struct session *to, struct session *from)
{  
   if( from->src_addr.type ){
      to->src_addr.type = from->src_addr.type;
      to->src_addr.name = strdup(from->src_addr.name);
   }
   to->src_addr.psm = from->src_addr.psm;
}

void free_addr(struct session *s)
{  
   if( s->src_addr.type ){
      s->src_addr.type = 0;
      free(s->src_addr.name);
   }
}

int free_session(void *d, void *u)
{
   struct session *s = d;

   free(s->name);   
   free_addr(s);
   
   llist_free(&s->up, free_cmd, NULL);   
   llist_free(&s->down, free_cmd, NULL);
	
   return 0;
}

void free_sessions(void)
{
   llist_free(&sessions, free_session, NULL);
}

/* 
 * Session search.
 * This functions can be called only once because they free sessions list.
 */ 
static int check_name(void *d, void *u)
{
   struct session *s = d;

   if(u && !strcmp(s->name, u))
      return 1;

   return 0;   
}

static int check_channel(void *d, void *u)
{
   struct session *s = d;
   int channel = (int)u;

   if(u && s->channel == channel )
      return 1;

   return 0;   
}

struct session * get_session_by_name(char *name)
{
   return (struct session *)llist_trav(&sessions, check_name, name);
}

struct session * get_session_by_channel(int channel)
{
   return (struct session *)llist_trav(&sessions, check_channel, (void *) channel);
}

/* 
 * Read config file. 
 */ 
int read_config(char *file) 
{
   static int cfg_loaded = 0;
   extern FILE *yyin;

   DEBUG_MARK();

   if( cfg_loaded ){
      free_sessions();
      syslog(LOG_INFO,"Reloading configuration file");
   }	 
   cfg_loaded = 1;

   DEBUG_MARK();

   llist_init(&sessions);

   DEBUG_MARK();

   if( !(yyin = fopen(file,"r")) ){
      syslog(LOG_ERR,"Can not open %s", file);
      return -1;      
   }

   DEBUG_MARK();

   yyparse();

   DEBUG_MARK();

   free_session(&default_session, NULL);

   fclose(yyin);
  
   DEBUG_MARK();

   return !llist_empty(&sessions);     
}
