/*
 * rfcomm.c -- Implementation of Bluetooth RFCOMM with TS 07.10, 
 *             Serial Port Emulation
 *
 * Copyright (C) 2000, 2001  Axis Communications AB
 *
 * Author: Mats Friden <mats.friden@axis.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Exceptionally, Axis Communications AB grants discretionary and
 * conditional permissions for additional use of the text contained
 * in the company's release of the AXIS OpenBT Stack under the
 * provisions set forth hereunder.
 *
 * Provided that, if you use the AXIS OpenBT Stack with other files,
 * that do not implement functionality as specified in the Bluetooth
 * System specification, to produce an executable, this does not by
 * itself cause the resulting executable to be covered by the GNU
 * General Public License. Your use of that executable is in no way
 * restricted on account of using the AXIS OpenBT Stack code with it.
 *
 * This exception does not however invalidate any other reasons why
 * the executable file might be covered by the provisions of the GNU
 * General Public License.
 *
 * $Id$
 *
 */

/****************** INCLUDE FILES SECTION ***********************************/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <syslog.h>

#include <asm/unaligned.h>
#include <asm/types.h>
#include <asm/byteorder.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/l2cap.h>

#include "rfcomm.h"
#include "rfcommd.h"
#include "lib.h"

#define MAX_TX_BUF_SIZE 2048

/****************** LOCAL FUNCTION DECLARATION SECTION **********************/

#define u8              __u8
#define u16             __u16
#define u32             __u32
#define s8              __s8
#define s16             __s16
#define s32             __s32
#define le16_to_cpu     __le16_to_cpu
#define cpu_to_le16     __cpu_to_le16

static void process_mcc(u8* data, u32 len, rfcomm_con *rfcomm, s32 long_pkt);
static s32 send_ua(rfcomm_con *rfcomm, u8 dlci);
static s32 send_dm(rfcomm_con *rfcomm, u8 dlci);
static s32 send_sabm(rfcomm_con *rfcomm, u8 dlci);
static s32 send_disc(rfcomm_con *rfcomm, u8 dlci);
static s32 send_uih(u8 *data, u32 len, rfcomm_con *rfcomm, u8 dlci);

static s32 send_pn_msg(rfcomm_con *rfcomm, u8 prior, u32 frame_size, u8 credit_flow, u8 credits, u8 dlci, u8 cr);
static s32 send_nsc_msg(rfcomm_con *rfcomm, mcc_type cmd, u8 cr);
static void set_uih_hdr(short_frame *uih_pkt, u8 dlci, u32 len, u8 cr);
static u32 crc_check(u8 *data, u32 length, u8 check_sum);
static u8 crc_calc(u8 *data, u32 length);
static void create_crctable(u8 table[]);
static rfcomm_con* get_new_rfcomm_con(void);
static rfcomm_con* get_rfcomm_con(u8 line);
static s32 get_connected_dlci(rfcomm_con *rfcomm);
static s32 rfcomm_send_credits(rfcomm_con *rfcomm, u8 dlci, u8 credits);
static s32 rfcomm_test_msg(rfcomm_con *rfcomm, u8 *test_pattern, u32 len, u8 cr);
static s32 rfcomm_fcon_msg(rfcomm_con *rfcomm, u8 cr);
static s32 rfcomm_fcoff_msg(rfcomm_con *rfcomm, u8 cr);
static s32 rfcomm_rpn_msg(rfcomm_con *rfcomm, u8 cr, u8 dlci, u8 req);
static s32 rfcomm_rls_msg(rfcomm_con *rfcomm, u8 cr, u8 dlci, u8 err_code);
static s32 rfcomm_msc_msg(rfcomm_con *rfcomm, u8 value, u8 cr, u8 dlci);
static s32 valid_dlci(u8 dlci);

/****************** GLOBAL VARIABLE DECLARATION SECTION *********************/

/****************** LOCAL VARIABLE DECLARATION SECTION **********************/

/* The crctable for the FEC field */
static u8 crctable[256];

/* One RFCOMM connection for each ttyBTx */
rfcomm_con rfcomm_con_list[BT_NBR_DATAPORTS];

rpn_values rpn_val;

/****************** FUNCTION DEFINITION SECTION *****************************/

/****************************************************************************
 * Functions reachable outside this file
 ****************************************************************************/

void rfcomm_reset_con(u8 line)
{	
	s32 j;

	/* check if valid line number */
	if (line >= BT_NBR_DATAPORTS)
	{
		D_ERR(FNC"invalid line!\n");
		return;
	}

	D_CTRL(FNC"line %d\n", line);

	rfcomm_con_list[line].line = line;
	rfcomm_con_list[line].credit_flow = 0;
	for (j = 0; j < 62; j++) {
		rfcomm_con_list[line].dlci[j].local_credits = 0;
		rfcomm_con_list[line].dlci[j].state = DISCONNECTED;
		rfcomm_con_list[line].dlci[j].mtu = DEF_RFCOMM_MTU;
		rfcomm_con_list[line].dlci[j].initiated = FALSE;
		rfcomm_con_list[line].dlci[j].initiator = FALSE;
	}
	rfcomm_con_list[line].initiator = FALSE;
	rfcomm_con_list[line].iff = NULL;
}

s32
rfcomm_module_init(void)
{
	return 0;
}

void
rfcomm_init(void)
{
	int i;

	DSYS("Initialising RFCOMM\n");
	create_crctable(crctable);
	
	/* Initiate rfcomm lines */
	for (i = 0; i < BT_NBR_DATAPORTS; i++) {
		rfcomm_reset_con(i);
	}

	/* Set the values in the rpn octets */
	rpn_val.bit_rate = 7; /* Irrelevent here */
	rpn_val.data_bits = 3; /* 8 databits */
	rpn_val.stop_bit = 0; /* One stopbit */
	rpn_val.parity = 0;
	rpn_val.parity_type = 0;
	rpn_val.res1 = 0;
	rpn_val.xon_input = 0;
	rpn_val.xon_output = 0;
	rpn_val.rtr_input = 0;
	rpn_val.rtr_output = 0;
	rpn_val.rtc_input = 0;
	rpn_val.rtc_output = 0;
	rpn_val.res2 = 0;
	rpn_val.xon = 0x11;
	rpn_val.xoff = 0x13;
	memset(&rpn_val.pm, 0 , 2); /* Set the mask to zero */
}

s32
rfcomm_set_mtu(rfcomm_con *rfcomm, u8 dlci, u32 new_mtu)
{
	DSYS(FNC"MTU set to:%d\n", new_mtu);
	if (rfcomm->dlci[dlci].state == DISCONNECTED) {
		rfcomm->dlci[dlci].mtu = new_mtu;
		return 0;
	} else {
		return -1;
	}
}

static char* state2name(u8 state)
{
	switch(state)
	{
	case DISCONNECTED :
		return "DISCONNECTED";
	case CONNECTING :
		return "CONNECTING";
	case NEGOTIATING :
		return "NEGOTIATING";
	case CONNECTED :
		return "CONNECTED";
	case DISCONNECTING :
		return "DISCONNECTING";
	default :
		return "Invalid State";
	}
}

/* This function should be called from the L2CAP layer data should pos32 at
   the beginning of the RFCOMM data, and len should be the length of that 
   data.*/

void 
rfcomm_receive_data(rfcomm_con *rfcomm, u8 *data, u32 len)
{
	short_frame *short_pkt; 
	long_frame *long_pkt;
	u8* uih_data_start;
	u32 uih_len;
	u8 tmp_dlci;

	D_REC(FNC"%d bytes\n", len);
	short_pkt = (short_frame*) data;

	tmp_dlci = (((short_pkt->h.addr.server_chn & 0x1f) << 1) |
		    (short_pkt->h.addr.d & 0x1));

	switch (CLR_PF(short_pkt->h.control)) {
	case SABM:
		D_CTRL(FNC"SABM-packet received\n");
    
		if (crc_check((u8*) short_pkt, LONG_CRC_CHECK, 
			      short_pkt->data[0]) < 0) {
			break;
		}

		/* If the SABM command wants to open the control channel, we
		   have to create a new rfcomm_con object */
		if ((tmp_dlci) == 0) {

			D_CTRL(FNC"server channel == 0\n");
      
			D_CTRL(FNC"setting l2cap_con, no serv.channel yet\n");
			rfcomm->dlci[0].state = CONNECTED;

                        D_CTRL(FNC"sending back UA - control channel\n");
                        send_ua(rfcomm, tmp_dlci);
		}
	
		/* If the channel isn't the control channel, we know that the
		   object exist. If we haven't opened any other channel than
		   the control channel we open the new channel here */

		else if (valid_dlci(tmp_dlci)) {
			int channel = short_pkt->h.addr.server_chn;
			int accept = 0;

			D_REC(FNC"setting server channel %d\n",
			      short_pkt->h.addr.server_chn);

			rfcomm->dlci[tmp_dlci].state = CONNECTED;

			if( rfcomm->iff->event )
				accept = (rfcomm->iff->event)(rfcomm->iff, CONNECT_IND, channel);

			if( accept ){
				D_CTRL(FNC"sending UA, dlci %d\n", tmp_dlci);
				send_ua(rfcomm, tmp_dlci);

				/* ---- DLCI connected. ---- */
				rfcomm->c_dlci = tmp_dlci;
				rfcomm->dlci[tmp_dlci].state = CONNECTED;

				(rfcomm->iff->event)(rfcomm->iff, CONNECT_CFM, channel);
			} else {
				D_CTRL(FNC"no server on channel %d, sending DM\n", tmp_dlci);
				send_dm(rfcomm, tmp_dlci);
			}
		} else {

			D_CTRL(FNC"no server on channel %d, sending DM\n",
			      tmp_dlci);
      
			send_dm(rfcomm, tmp_dlci);
		}

		break;
    
	case UA:
		D_CTRL(FNC"UA packet received on line %d\n", rfcomm->line);

		if (rfcomm->dlci[0].state == CONNECTING) {
			/* Now we are the initiating side and we have got a
			   response to our SABM command. We then send a PN
			   messages to negotiat the frame length */
			rfcomm->dlci[0].state = CONNECTED;
			tmp_dlci = ((rfcomm->server_chn << 1) |
				    ((~rfcomm->initiator) & 0x1));
			
			/* must fit in l2cap mtu incl rfcomm hdrs */
			send_pn_msg(rfcomm, 7, rfcomm->dlci[tmp_dlci].mtu, 0xf,
				    START_CREDITS, tmp_dlci, TRUE);
			rfcomm->dlci[tmp_dlci].state = NEGOTIATING;

		} else if (rfcomm->dlci[tmp_dlci].state == NEGOTIATING) {
			rfcomm->c_dlci = tmp_dlci;
			rfcomm->dlci[tmp_dlci].state = CONNECTED;

			D_CTRL(FNC"successfully connected on DLCI:%d\n",
			       tmp_dlci);
			
			/* FIXME: Should we always do this ? */
			rfcomm_msc_msg(rfcomm, EA | RTC | RTR | DV, MCC_CMD,
				       tmp_dlci);
			rfcomm->dlci[tmp_dlci].initiator = TRUE;
			/* ---- DLCI connected. ---- */
			if( rfcomm->iff->event )
				(rfcomm->iff->event)(rfcomm->iff, CONNECT_CFM, rfcomm->server_chn);
		} else if (rfcomm->dlci[tmp_dlci].state == DISCONNECTING) {
		  
			if (tmp_dlci == 0) {
				/* Control channel */
				DSYS(FNC"RFCOMM disconnected ctrl ch (local) on line [%d]\n", rfcomm->line);
				rfcomm->dlci[0].state = DISCONNECTED;
				close(rfcomm->iff->sock);
			} else {
				/* Data channel */
				s32 tmp;
				rfcomm->dlci[tmp_dlci].state = DISCONNECTED;
				tmp = get_connected_dlci(rfcomm);

				if (tmp >= 0) {
					rfcomm->dlci[tmp].state = DISCONNECTING;
					send_disc(rfcomm, tmp);
					/* ---- DLCI disconnected. ---- */
					if( rfcomm->iff->event )
						(rfcomm->iff->event)(rfcomm->iff, DISCONN_CFM, rfcomm->server_chn);
				} else {
					D_ERR("Could not find connected DLCI\n");
				}
				
			}
		} else if (rfcomm->dlci[tmp_dlci].state == DISCONNECTED) {
			rfcomm->dlci[tmp_dlci].state = CONNECTED;
		} else {
			D_WARN(FNC" Something wrong receiving UA packet\n");
		}    
		break;
    
		/* Disconnect mode, 'NAK on SABM/DISC' */
	case DM:
		D_CTRL(FNC"DM packet received\n");
		rfcomm->dlci[tmp_dlci].state = DISCONNECTED;
		/* ---- DLCI disconnected. ---- */
		if( rfcomm->iff->event )
			(rfcomm->iff->event)(rfcomm->iff, DISCONN_CFM, rfcomm->server_chn);

		close(rfcomm->iff->sock);
		close(rfcomm->iff->pty);
		break;
    
	case DISC:
		D_CTRL(FNC"DISC packet received\n");
		if (crc_check(data, LONG_CRC_CHECK, short_pkt->data[0]) < 0) {
			break;
		}
		/* When the serverchannel is closing the whole connection
		   should be removed */
		if (rfcomm->dlci[tmp_dlci].state == DISCONNECTED) {
			send_dm(rfcomm, tmp_dlci);
		} else if ((short_pkt->h.addr.server_chn) == 0) {
			rfcomm->dlci[0].state = DISCONNECTED;

			DSYS("RFCOMM control ch disconnected (remotely) [line:%d]\n", 
			       rfcomm->line);
			send_ua(rfcomm, 0);
			close(rfcomm->iff->pty);
		} else {
			rfcomm->dlci[tmp_dlci].state = DISCONNECTED;
			send_ua(rfcomm, tmp_dlci);
			close(rfcomm->iff->pty);
			D_CTRL("dlci %d was disconnected\n", tmp_dlci);
		}
		break;
    
	case UIH:
		if ((short_pkt->h.length.ea) == 0) {
			/* Then we cast the rfcomm packet to a long rfcomm
			   packet */ 
			D_REC(FNC"Long UIH packet received\n");
			long_pkt = (long_frame*) data;
			swap_long_frame(long_pkt);
			uih_len = long_pkt->h.length.bits.len;
			uih_data_start = long_pkt->h.data;
			D_REC(FNC"long packet length %d\n",
			      uih_len);
			if (uih_len > (len - 5)) {
				D_WARN(FNC", Long packet length doesn't match,"
				"setting length to l2cap len %d - 5\n", len);
				uih_len = len - 5;
			}
			
		} else {
			D_REC(FNC"Short UIH pkt received\n");
			uih_len = short_pkt->h.length.len;
			uih_data_start = short_pkt->data;
			if (uih_len > (len - 4)) {
				D_WARN(FNC", Short packet length doesn't "
				"match, setting length to l2cap len %d - 4\n", 
				len);
				uih_len = len - 4;
			}
		}

		if (GET_PF(short_pkt->h.control)) {
			D_REC(FNC" %d more credits on dlci:%d...\n", 
			       *uih_data_start, tmp_dlci);

			rfcomm->dlci[tmp_dlci].local_credits += uih_data_start[0];
			D_REC(FNC"Local_credits:%d\n", rfcomm->dlci[tmp_dlci].local_credits); 
			uih_data_start++;
			if (uih_len == 0) {
				break;
			}
			
			/* feed uih data to tty if any */

		}

		if (crc_check(data, SHORT_CRC_CHECK,
			      *(uih_data_start + uih_len)) < 0) {
			break;
		}
    
		if (short_pkt->h.addr.server_chn == 0) {
			D_REC(FNC"UIH on serv_channel 0\n");
			process_mcc(data, len, rfcomm,
				    !(short_pkt->h.length.ea));
		} else {
			if (rfcomm->credit_flow) {
				if(rfcomm->dlci[tmp_dlci].remote_credits) {
					--rfcomm->dlci[tmp_dlci].remote_credits;
				}
				
				D_CTRL(FNC": Remote credits: %d\n",rfcomm->dlci[tmp_dlci].remote_credits);
				/* Send new credits in packet only if we totally run out of them. When we send
				   data to the other side we usually send credits needed together with the data
				   (piggy-backed) */
				if (rfcomm->dlci[tmp_dlci].remote_credits == 0) {
					u8 newcredits = MAX_CREDITS - rfcomm->dlci[tmp_dlci].remote_credits;
					rfcomm->dlci[tmp_dlci].remote_credits += newcredits;
					rfcomm_send_credits(rfcomm, tmp_dlci, newcredits);
					D_SND(FNC"Remote credits: %d\n",rfcomm->dlci[tmp_dlci].remote_credits);
					
				}
			}
			if (uih_len)
				write_n(rfcomm->iff->pty, uih_data_start, uih_len);
		}
		break;
    
	default:
		D_REC(FNC"illegal packet\n");
		break;
	}

}


/* Copies the tty data into the stack and creates headers and FCS, then the
   packet is sent to the lower layer L2CAP. The packet must not be larger
   than the actual MTU for the L2CAP layer. */

s32 
rfcomm_send_data(rfcomm_con *rfcomm, u8 *data, u32 count)
{
	u32 c, c_framed;
	u32 total = 0;
	u8 dlci;

	dlci = rfcomm->c_dlci;
	
	if (dlci == 0) {
		D_ERR(FNC"Not allowed to send data on DLCI 0\n");
		return -1;
	}
	
	if (rfcomm == NULL) {
		D_ERR(FNC" ERROR rfcomm_con == NULL\n");
		return -1;
	} else if(rfcomm->dlci[0].state == FLOW_STOPPED) {
		D_SND(FNC"Flow stopped on all channels, returning zero\n");
		return 0;
	} else if (rfcomm->dlci[dlci].state == FLOW_STOPPED) {
		D_SND(FNC"Flow stopped on DLCI %d, returning zero\n", dlci);
		return 0;
	}
	/* Check whether there are any data channels connected */
	else if (rfcomm->dlci[dlci].state == CONNECTED) {
		D_SND(FNC"trying to send %d bytes\n", count);
		while (count) {	
		  if (rfcomm->credit_flow &&
		     (rfcomm->dlci[dlci].local_credits <= 0)) {
		    D_SYS(FNC"Flow stopped, no credits (%d) on "
			"DLCI %d, returning %d\n", 
			rfcomm->dlci[dlci].local_credits, dlci, total);
		    /* "It is always allowed to send frames containing 
		        no user data (length field = 0) when credit based 
			flow control is in use." 
			Bluetooth Spec 1.1, p.418
		     */
		    /* send_uih(data, 0, rfcomm, dlci); */
		    return total;
		  }
			
		  /* As long as there are anything to send we try to send
		     it. We split the data into parts not bigger than the
		     current MTU size. */
			
		  if (rfcomm->dlci[dlci].mtu) {
		  	c = MIN(count, rfcomm->dlci[dlci].mtu);
		  } else {
			c = count;
		  }
			
		  /* FIXME: Optimize c! If we try to send 200 bytes and
		     there are only 150 bytes space in buffer, set c to
		     150 - sizeof(rfcomm_tx_buf)
		  */
		  c_framed = send_uih(data, c, rfcomm, dlci);

		  if (c_framed < 0) {
		    return c_framed; /* error */
		  } else if (c_framed == 0) {
		    return total;
		  } else {
		    if (rfcomm->credit_flow) {
		      --rfcomm->dlci[dlci].local_credits;
		      D_SND(FNC"Local credits:%d\n", 
		      rfcomm->dlci[dlci].local_credits);
		    }

		    total += c;
		    data += c;
		    count -= c;
		  }
		}
		D_SND(FNC"sent %d bytes\n", total);
		return total;
	} else {
		D_ERR(FNC"DLCI %d not connected\n", dlci);
		return -1;
	}

}

s32
rfcomm_flow_stop(u8 line, u8 dlci)
{
	if (rfcomm_con_list[line].credit_flow &&(rfcomm_con_list[line].dlci[dlci].local_credits <= 0)) {
		return TRUE;
	} else if (rfcomm_con_list[line].dlci[dlci].state == FLOW_STOPPED) {
		return TRUE;
	} else if (rfcomm_con_list[line].dlci[0].state == FLOW_STOPPED) {
		return TRUE;
	} else {
		return FALSE;
	}
}

void 
rfcomm_send_testdata(u32 count, u8 line)
{
	static u8 testdata[1024];
	static s32 testdata_init = 0;
	s32 i;

	DSYS("rfcomm_send_testdata: sending %d bytes\n",count);
	if (!testdata_init) {
		for (i = 0; i < 1024; i++) {
			testdata[i] = (i%25)+ 65;
		}
	} 
	rfcomm_send_data(&rfcomm_con_list[line], testdata, count);
}

/* Parses a multiplexer control channel packet */
void 
process_mcc(u8* data, u32 len, rfcomm_con *rfcomm, s32 longpkt)
{
	mcc_short_frame *mcc_short_pkt;
    
	D_CTRL("process_mcc\n");

	if (longpkt) {
		mcc_short_pkt = (mcc_short_frame*)(((long_frame*)data)->data);
	} else {
		mcc_short_pkt = (mcc_short_frame*)(((short_frame*)data)->data);
	}
  
	switch (mcc_short_pkt->h.type.type) {
	case TEST:
		if (mcc_short_pkt->h.type.cr == MCC_RSP) {
			D_CTRL(FNC"Received test command\n");
		} else {
			if ((mcc_short_pkt->h.length.ea) == 0) {
				mcc_long_frame *mcc_long_pkt;
				mcc_long_pkt = (mcc_long_frame*) mcc_short_pkt;
				swap_mcc_long_frame(mcc_long_pkt);
				rfcomm_test_msg(rfcomm, mcc_long_pkt->value,
					      mcc_long_pkt->h.length.bits.len,
					      MCC_RSP);
			} else {
				rfcomm_test_msg(rfcomm, mcc_short_pkt->value, 
					      mcc_short_pkt->h.length.len,
					      MCC_RSP);
			}
		}
		break;
    
	case FCON:  /* Flow control on command */
		D_CTRL(FNC"Received Flow control on command\n");
		if (mcc_short_pkt->h.type.cr == MCC_CMD) {
			rfcomm->dlci[0].state = CONNECTED;
			rfcomm_fcon_msg(rfcomm, MCC_RSP);
		}
		break;
    
	case FCOFF:  /* Flow control off command */
		D_CTRL(FNC"Received Flow control off command\n");
		if (mcc_short_pkt->h.type.cr == MCC_CMD) {
			rfcomm->dlci[0].state = FLOW_STOPPED;
			rfcomm_fcoff_msg(rfcomm, MCC_RSP);
		}
		break;
    
	case MSC:  /* Modem status command */
	{
		u8 dlci;
		u8 v24_sigs;
		
		dlci = (mcc_short_pkt->value[0]) >> 2;
		v24_sigs = mcc_short_pkt->value[1];

		if (rfcomm->dlci[dlci].state == DISCONNECTED) {
			send_dm(rfcomm, dlci);
			break;
		}
		if (mcc_short_pkt->h.type.cr == MCC_CMD) {
			D_CTRL(FNC"Received Modem status command\n");
			if (v24_sigs & 2) {
				D_CTRL(FNC"Device unable to accept frames\n");
				rfcomm->dlci[dlci].state = FLOW_STOPPED;
			} else {
				rfcomm->dlci[dlci].state = CONNECTED;
				D_CTRL(FNC"Flow ON, dlci %d\n", dlci);
  			}

			rfcomm_msc_msg(rfcomm, v24_sigs, MCC_RSP, dlci);
			
			if (!(rfcomm->dlci[dlci].initiated) && !(rfcomm->dlci[dlci].initiator)) {
				rfcomm_msc_msg(rfcomm, EA | RTR | RTC | DV,
					       MCC_CMD, dlci);
				rfcomm->dlci[dlci].initiated = TRUE;
			}
		} else {
			D_CTRL(FNC"Received Modem status response\n");

			if (v24_sigs & 2) {
				D_CTRL(FNC"Flow stop accepted\n");
			}
		}
		break;
	}
    
	case RPN:  /* Remote port negotiation command */
	{
		u8 tmp_dlci;

		tmp_dlci = (mcc_short_pkt->value[0]) >> 2;
		
		if (mcc_short_pkt->h.type.cr == MCC_CMD) {
			if (mcc_short_pkt->h.length.len == 1) {
				D_CTRL(FNC"Received Remote port negotiation command\n");
				rfcomm_rpn_msg(rfcomm, MCC_RSP, tmp_dlci, 0);
			} else {
				/* Accept the other sides settings (accept all
				   for now) */
				D_CTRL(FNC"Received Remote port negotiation response\n");
				memcpy(&rpn_val, &mcc_short_pkt->value[1], 8);
				rfcomm_rpn_msg(rfcomm, MCC_RSP, tmp_dlci, 0);
				/* Zero the parametermask after response */
				memset(&rpn_val.pm, 0, sizeof(rpn_val.pm));
			}
		}	
		break;
	}

	case RLS: /* Remote line status */
	{
		u8 tmp_dlci;
		u8 err_code;
			
		D_CTRL(FNC"Received Remote line status\n");
		if (mcc_short_pkt->h.type.cr == MCC_CMD) {
			tmp_dlci = mcc_short_pkt->value[0] >> 2;
			err_code = mcc_short_pkt->value[1];
		
			rfcomm_rls_msg(rfcomm, MCC_RSP, tmp_dlci, err_code);
		}
		break;
	}

	case PN:  /* DLC parameter negotiation */
	{
		pn_msg *pn_pkt;

		D_CTRL(FNC"Received DLC parameter negotiation, PN\n");
		if (longpkt) {
			/* If a long length field is used, then move the
			   payload pointer one byte ahead */
			data++;
		}  
		pn_pkt = (pn_msg*) data;

		if (pn_pkt->mcc_s_head.type.cr == MCC_CMD) {
			u8 tmp_dlci;
			u16 frame_size;
			u8 credit, nbrof_credits;
			
			/* If the incoming messages is a command we send
			   back a response */

			tmp_dlci = pn_pkt->dlci;
			frame_size = le16_to_cpu(pn_pkt->frame_size);

			frame_size = MIN(frame_size, rfcomm->local_mtu - RFCOMM_MAX_HDR_SIZE);
			credit = pn_pkt->credit_flow;
			nbrof_credits = pn_pkt->credits;
			if (credit) {
				DSYS(FNC"Using credit flow control, initiating credits are %d on dlci %d\n", nbrof_credits, tmp_dlci);

				rfcomm->credit_flow = TRUE;
				rfcomm->dlci[tmp_dlci].local_credits = nbrof_credits;
				D_CTRL(FNC"Local credits:%d\n", rfcomm->dlci[tmp_dlci].local_credits); 
				rfcomm->dlci[tmp_dlci].remote_credits = START_CREDITS;
				D_CTRL(FNC"Remote credits: %d\n",rfcomm->dlci[tmp_dlci].remote_credits);
				send_pn_msg(rfcomm, pn_pkt->prior, frame_size,
					    credit ^ 1, START_CREDITS, 
					    tmp_dlci, MCC_RSP);
			} else {
				send_pn_msg(rfcomm, pn_pkt->prior, frame_size,
					    0, 0, tmp_dlci, MCC_RSP);
			}
			rfcomm->dlci[tmp_dlci].mtu = frame_size;
			D_CTRL(FNC"process_mcc : mtu set to %d\n", 
			      rfcomm->dlci[tmp_dlci].mtu);
		} else {
			u8 tmp_dlci;
			tmp_dlci = pn_pkt->dlci;
			D_CTRL(FNC"received PN response with:\n");
			D_CTRL(FNC"Frame size:%d\n", le16_to_cpu(pn_pkt->frame_size));
			D_CTRL(FNC"credit_flow:%d\n", pn_pkt->credit_flow);
			D_CTRL(FNC"credits:%d\n", pn_pkt->credits);
			
			rfcomm->dlci[tmp_dlci].mtu  = le16_to_cpu(pn_pkt->frame_size);
			
			if (pn_pkt->credit_flow == 0xe) {
				DSYS(FNC"Credit flow control used\n");
				rfcomm->credit_flow = TRUE;
				rfcomm->dlci[tmp_dlci].local_credits = pn_pkt->credits;
				D_CTRL(FNC"Local credits:%d\n", rfcomm->dlci[tmp_dlci].local_credits); 
			}

			D_CTRL(FNC"process_mcc : mtu set on dlci:%d to %d\n", 
			       tmp_dlci, rfcomm->dlci[tmp_dlci].mtu);

			if (rfcomm->dlci[tmp_dlci].state == NEGOTIATING) {
				send_sabm(rfcomm, tmp_dlci);
			}
		}
		break;
	}
    
	case NSC: /* Non supported command resonse */
		D_CTRL(FNC"Received Non supported command response\n");
		break;
    
	default:  /* Non supported command received */
		D_REC(FNC"Received a non supported command\n");
    
		send_nsc_msg(rfcomm, mcc_short_pkt->h.type, MCC_RSP);
    
		break;
	}
}


/* Sending packet functions */

/* Creates a UA packet and puts it at the beginning of the pkt pointer */
                         
s32 
send_ua(rfcomm_con *rfcomm, u8 dlci)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	short_frame *ua_pkt;
	u32 rfcomm_frame_size;
  
	D_CTRL("send_ua: Creating UA packet to DLCI %d\n",dlci);

	rfcomm_frame_size = sizeof(short_frame) + FCS_SIZE;
	ua_pkt = (short_frame*) &tx_buf;
  
	/* Always one in the address field */
	ua_pkt->h.addr.ea = 1;
	/* If we are the responder the cr bit should be set in a response */
	ua_pkt->h.addr.cr = ((~(rfcomm->initiator)) & 0x1);
	/* The control channel is on DLCI 0, i.e. d-bit should not be set */
	ua_pkt->h.addr.d = (dlci) & 0x1;
	ua_pkt->h.addr.server_chn = (dlci) >> 0x1;
	/* In RFCOMM the F-bit should always be set when sending UA packets */
	ua_pkt->h.control = SET_PF(UA);
	/* The UA packet has length zero, and then the ea bit in the length
	   field should  be set */
	ua_pkt->h.length.ea = 1;
	ua_pkt->h.length.len = 0;
	ua_pkt->data[0] = crc_calc((u8*) ua_pkt, LONG_CRC_CHECK);
  
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

/* Creates a DM packet and puts it at the beginning of the pkt pointer */

s32 
send_dm(rfcomm_con *rfcomm, u8 dlci)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	short_frame *dm_pkt;
	u32 rfcomm_frame_size;
  
	D_CTRL("send_dm: Creating DM packet to DLCI %d\n",dlci);
  
	rfcomm_frame_size = sizeof(short_frame) + FCS_SIZE;
	dm_pkt = (short_frame*) &tx_buf;

	/* Always one in the address field */
	dm_pkt->h.addr.ea = 1;
	/* If we are the responder the cr bit should be set in a response */
	dm_pkt->h.addr.cr =  ((~(rfcomm->initiator)) & 0x1);
	/* The control channel is on DLCI 0, i.e. d-bit should not be set */
	dm_pkt->h.addr.d = dlci & 0x1;
	dm_pkt->h.addr.server_chn = dlci >> 0x1;
	/* In RFCOMM the F-bit should always be set when sending DM packets */
	dm_pkt->h.control = SET_PF(DM);
	/* The DM packet has length zero, and then the ea bit in the length
	   field should  be set */
	dm_pkt->h.length.ea = 1;
	dm_pkt->h.length.len = 0;
	dm_pkt->data[0] = crc_calc((u8*) dm_pkt, LONG_CRC_CHECK);

	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

s32 
send_sabm(rfcomm_con *rfcomm, u8 dlci)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	short_frame *sabm_pkt;
	u32 rfcomm_frame_size;

	D_CTRL("send_sabm: Creating SABM packet to DLCI %d\n",dlci);

	rfcomm_frame_size = sizeof(short_frame) + FCS_SIZE;
	sabm_pkt = (short_frame*) &tx_buf;
  
	/* Always one in the address field */
	sabm_pkt->h.addr.ea = 1;
	/* If we are the responder the cr bit should be set in a response */
	sabm_pkt->h.addr.cr = ((rfcomm->initiator) & 0x1);
	sabm_pkt->h.addr.d = dlci & 0x1;
	sabm_pkt->h.addr.server_chn = dlci >> 0x1;
	/* In RFCOMM the F-bit should always be set when sending SABM packets
	 */
	sabm_pkt->h.control = SET_PF(SABM);
	/* The SABM packet has length zero, and then the ea bit in the length
	   field should  be set */
	sabm_pkt->h.length.ea = 1;
	sabm_pkt->h.length.len = 0;
	sabm_pkt->data[0] = crc_calc((u8*) sabm_pkt, LONG_CRC_CHECK);
  
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

s32 
send_disc(rfcomm_con *rfcomm, u8 dlci)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	short_frame *disc_pkt;
	u32 rfcomm_frame_size;

	D_CTRL("send_disc: Creating DISC packet to DLCI %d\n",dlci);

	rfcomm_frame_size = sizeof(short_frame) + FCS_SIZE;
	disc_pkt = (short_frame*) &tx_buf;
  
	/* Always one in the address field */
	disc_pkt->h.addr.ea = 1;
	/* If we are the responder the cr bit should be set in a response */
	disc_pkt->h.addr.cr = ((rfcomm->initiator) & 0x1);
	disc_pkt->h.addr.d = dlci & 0x1;
	disc_pkt->h.addr.server_chn = dlci >> 0x1;
	/* In RFCOMM the F-bit should always be set when sending DISC packets
	 */
	disc_pkt->h.control = SET_PF(DISC);
	/* The DISC packet has length zero, and then the ea bit in the length
	   field should  be set */
	disc_pkt->h.length.ea = 1;
	disc_pkt->h.length.len = 0;
	disc_pkt->data[0] = crc_calc((u8*) disc_pkt, LONG_CRC_CHECK);
  
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
} 


/* Creates a UIH packet and puts it at the beginning of the pkt pointer
   The function returns how many bytes the UIH packet consist of */

s32 
send_uih(u8 *data, u32 len, rfcomm_con *rfcomm, u8 dlci)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	u32 rfcomm_frame_size;
	u8  send_credit = 0;
	s32 retval = 0;
	u8 newcredits = 0;
	int max_tx_buf_size;
	int header_size;

	D_CTRL(FNC"Creating UIH packet with %d bytes data to DLCI %d\n",
	       len, dlci);

	header_size = (len > SHORT_PAYLOAD_SIZE) ? 
		sizeof(long_frame) : sizeof(short_frame);

	if (rfcomm->dlci[dlci].mtu) {
		/* The following estimate includes the possible
		   credit field as part of the negotiated maximum 
		   payload size.  The payload length field value won't 
		   include the credit field.  Inserting the credit 
		   field doesn't affect the allowed maximum size.  
		   This is aimed at keeping the L2CAP payload constant.  
		   See pp.418-419 of Bluetooth Specification 1.1
		   and ETSI 07.10 at
		     http://pda.etsi.org/exchangefolder/ts_101369v070200p.pdf
		 */
		max_tx_buf_size = header_size + 
			rfcomm->dlci[dlci].mtu + FCS_SIZE;
		if (max_tx_buf_size > MAX_TX_BUF_SIZE) {
			max_tx_buf_size = MAX_TX_BUF_SIZE;
		}
	} else {
		max_tx_buf_size = MAX_TX_BUF_SIZE;
	}
	
	if (rfcomm->credit_flow 
	    && (rfcomm->dlci[dlci].remote_credits < MIN_CREDITS)) {
		D_SND(FNC"Sending more credits to remote port\n");
		send_credit = 1;
		newcredits = MAX_CREDITS - rfcomm->dlci[dlci].remote_credits;
		rfcomm->dlci[dlci].remote_credits += newcredits;
	}

	rfcomm_frame_size = header_size + send_credit + len + FCS_SIZE;
	if (rfcomm_frame_size > max_tx_buf_size) {
		u32 orig_len = len;
		len = max_tx_buf_size - header_size - FCS_SIZE - send_credit;
		header_size = (len > SHORT_PAYLOAD_SIZE) ? 
			sizeof(long_frame) : sizeof(short_frame);
		rfcomm_frame_size = max_tx_buf_size;
		D_SYS(FNC
		  "Cut the UIH data size from %d to %d on DLCI %d\n", 
		  orig_len, len, dlci);
	}

	if (len > SHORT_PAYLOAD_SIZE) {
		long_frame *l_pkt;
		l_pkt = (long_frame*) &tx_buf;
		set_uih_hdr((void*) l_pkt, dlci, len, rfcomm->initiator);
		if (send_credit) {
			l_pkt->h.control = SET_PF(UIH);
			l_pkt->data[0] = (newcredits);
			D_SND(FNC"Remote credits: %d\n", 
				rfcomm->dlci[dlci].remote_credits);
			memcpy(l_pkt->data + 1, data, len);
		} else {
			memcpy(l_pkt->data, data, len);
		}
		l_pkt->data[len + send_credit] = crc_calc((u8*) l_pkt, SHORT_CRC_CHECK);
		swap_long_frame(l_pkt);
	} else {
		short_frame *s_pkt;
		s_pkt = (short_frame*) &tx_buf;
		/* Always one */
		set_uih_hdr((void*) s_pkt, dlci, len, rfcomm->initiator);
		if (send_credit) {
			s_pkt->h.control = SET_PF(UIH);
			s_pkt->data[0] = (newcredits);
			D_SND(FNC"Remote credits: %d\n",
				rfcomm->dlci[dlci].remote_credits);
			memcpy(s_pkt->data + 1, data, len);
		} else {
			memcpy(s_pkt->data, data, len);
		}
		s_pkt->data[len + send_credit] = crc_calc((u8*) s_pkt, SHORT_CRC_CHECK);
	}

	/* FIXME - How should we propagate result up to higher layers ?
	   through len or success/no success? */
	
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

s32
rfcomm_send_credits(rfcomm_con *rfcomm, u8 dlci, u8 credits)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	short_frame *uih_pkt;
	u32 rfcomm_frame_size;
  
	D_CTRL(FNC"give %d credits to dlci %d\n", credits, dlci);

	rfcomm_frame_size = (sizeof(short_frame) + 1 + FCS_SIZE);

	uih_pkt = (short_frame*) &tx_buf;
	set_uih_hdr(uih_pkt, dlci, 0, (~rfcomm->initiator)^1);
	uih_pkt->h.control = SET_PF(UIH);

	uih_pkt->data[0] = credits;
	uih_pkt->data[1] = crc_calc((u8*) uih_pkt, SHORT_CRC_CHECK);
	
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}


/* Sends Multiplexer command packets funktions */

/* Sends a test messages with a specific test pattern. */

s32 
rfcomm_test_msg(rfcomm_con *rfcomm, u8 *test_pattern, u32 len, u8 cr)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	u32 rfcomm_frame_size;
   
	if (len > SHORT_PAYLOAD_SIZE) {
		long_frame *uih_pkt;
		mcc_long_frame *mcc_pkt;
    
		/* Create long uih packet and long mcc packet */
		rfcomm_frame_size = (sizeof(long_frame)
				     + sizeof(mcc_long_frame) + len +FCS_SIZE);
		if (rfcomm_frame_size > MAX_TX_BUF_SIZE) {
			len = MAX_TX_BUF_SIZE - sizeof(long_frame) 
				- sizeof(mcc_long_frame) - FCS_SIZE;
			rfcomm_frame_size = MAX_TX_BUF_SIZE;
		}  /* FIXME: is it legal to truncate outgoing RFCOMM data? -- IL */

		uih_pkt = (long_frame*) &tx_buf;
    
		set_uih_hdr((short_frame*) uih_pkt, CTRL_CHAN, len + 
			    sizeof(mcc_long_frame), rfcomm->initiator);
		uih_pkt->data[uih_pkt->h.length.bits.len] = crc_calc((u8*) uih_pkt, 
								SHORT_CRC_CHECK);
		mcc_pkt = (mcc_long_frame*) uih_pkt->data;

		/* Always one in the TEST messages */
		mcc_pkt->h.type.ea = EA;
		/* cr tells whether it is a commmand (1) or a response (0) */
		mcc_pkt->h.type.cr = cr;
		mcc_pkt->h.type.type = TEST;
		mcc_pkt->h.length.bits.ea = EA;
		mcc_pkt->h.length.bits.len = len;
		swap_long_frame(uih_pkt);
		swap_mcc_long_frame(mcc_pkt);
		memcpy(mcc_pkt->value,test_pattern,len);
	} else if (len > (SHORT_PAYLOAD_SIZE-sizeof(mcc_short_frame))) {
		long_frame *uih_pkt;
		mcc_short_frame *mcc_pkt;
    
		/* Create long uih packet and short mcc packet */
		rfcomm_frame_size = (sizeof(long_frame)
				     + sizeof(mcc_short_frame) + len+FCS_SIZE);
		uih_pkt = (long_frame*) &tx_buf;
    
		set_uih_hdr((short_frame*) uih_pkt, CTRL_CHAN, len +
			    sizeof(mcc_short_frame), rfcomm->initiator);
		uih_pkt->data[uih_pkt->h.length.bits.len] = crc_calc((u8*) uih_pkt, 
								SHORT_CRC_CHECK);
		mcc_pkt = (mcc_short_frame*) uih_pkt->data;

		/* Always one in the TEST messages */
		mcc_pkt->h.type.ea = EA;
		/* cr tells whether it is a commmand (1) or a response (0) */
		mcc_pkt->h.type.cr = cr;
		mcc_pkt->h.type.type = TEST;
		mcc_pkt->h.length.ea = EA;
		mcc_pkt->h. length.len = len;
		swap_long_frame(uih_pkt);
		memcpy(mcc_pkt->value,test_pattern,len);
	} else {
		short_frame *uih_pkt;
		mcc_short_frame *mcc_pkt;

		/* Creat short uih packet and short mcc packet */
		rfcomm_frame_size = (sizeof(short_frame)
				     + sizeof(mcc_short_frame) + len+FCS_SIZE);
		uih_pkt = (short_frame*) &tx_buf;
    
		set_uih_hdr((void*) uih_pkt, CTRL_CHAN, len
			    + sizeof(mcc_short_frame), rfcomm->initiator);
		uih_pkt->data[uih_pkt->h.length.len] = crc_calc((u8*) uih_pkt, 
								SHORT_CRC_CHECK);
		mcc_pkt = (mcc_short_frame*) uih_pkt->data;
    
		/* Always one in the TEST messages */
		mcc_pkt->h.type.ea = EA;
		/* cr tells whether it is a commmand (1) or a response (0) */
		mcc_pkt->h.type.cr = cr;
		mcc_pkt->h.type.type = TEST;
		mcc_pkt->h.length.ea = EA;
		mcc_pkt->h. length.len = len;
		memcpy(mcc_pkt->value,test_pattern,len);
    
	}
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

/* Turns on the RFCOMM flow control */

s32 
rfcomm_fcon_msg(rfcomm_con *rfcomm, u8 cr)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	short_frame *uih_pkt;
	mcc_short_frame *mcc_pkt;
	u32 rfcomm_frame_size;
  
	rfcomm_frame_size = (sizeof(short_frame) + sizeof(mcc_short_frame)
			     + FCS_SIZE);
	uih_pkt = (short_frame*) &tx_buf;
	set_uih_hdr(uih_pkt, CTRL_CHAN, sizeof(mcc_short_frame),
		    rfcomm->initiator);
	uih_pkt->data[sizeof(mcc_short_frame)] = crc_calc((u8*) uih_pkt, 
							  SHORT_CRC_CHECK);
	mcc_pkt = (mcc_short_frame*) (uih_pkt->data);
  
	mcc_pkt->h.type.ea = EA;
	mcc_pkt->h.type.cr = cr;
	mcc_pkt->h.type.type = FCON;
	mcc_pkt->h.length.ea = EA;
	mcc_pkt->h.length.len = 0;
  
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

/* Turns off the RFCOMM flow control */

s32 
rfcomm_fcoff_msg(rfcomm_con *rfcomm, u8 cr)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	short_frame *uih_pkt;
	mcc_short_frame *mcc_pkt;
	u32 rfcomm_frame_size;
  
	rfcomm_frame_size = (sizeof(short_frame) + sizeof(mcc_short_frame)
			     + FCS_SIZE);
	uih_pkt = (short_frame*) &tx_buf;
	set_uih_hdr(uih_pkt, CTRL_CHAN, sizeof(mcc_short_frame),
		    rfcomm->initiator);
	uih_pkt->data[sizeof(mcc_short_frame)] = crc_calc((u8*) uih_pkt, 
							  SHORT_CRC_CHECK);
	mcc_pkt = (mcc_short_frame*) (uih_pkt->data);

	mcc_pkt->h.type.ea = 1;
	mcc_pkt->h.type.cr = cr;
	mcc_pkt->h.type.type = FCOFF;
	mcc_pkt->h.length.ea = 1;
	mcc_pkt->h.length.len = 0;

	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

s32
rfcomm_rpn_msg(rfcomm_con *rfcomm, u8 cr, u8 dlci, u8 req)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	rpn_msg* rpn_pkt;
	u32 rfcomm_frame_size;
	u32 rfcomm_payload_size;

	rfcomm_frame_size = sizeof(rpn_msg);

	if (req) {
		rfcomm_frame_size -= sizeof(rpn_values);
	}
	
	rfcomm_payload_size = (rfcomm_frame_size - sizeof(short_frame)
			       - FCS_SIZE);
	
	rpn_pkt = (rpn_msg*) &tx_buf;
  
	set_uih_hdr((short_frame*) rpn_pkt, CTRL_CHAN, rfcomm_payload_size,
		    rfcomm->initiator);

	rpn_pkt->fcs = crc_calc((u8*) rpn_pkt, SHORT_CRC_CHECK);

	rpn_pkt->mcc_s_head.type.ea = EA;
	rpn_pkt->mcc_s_head.type.cr = cr;
	rpn_pkt->mcc_s_head.type.type = RPN;
	rpn_pkt->mcc_s_head.length.ea = EA;

	rpn_pkt->dlci.ea = EA;
	rpn_pkt->dlci.cr = 1; /* Fixed value */
	rpn_pkt->dlci.d = dlci & 1;
	rpn_pkt->dlci.server_chn = (dlci >> 1);
	
	if (req) {
		rpn_pkt->mcc_s_head.length.len = 1;
		/* Fix, since the packet is ends here when it is a request */
		rpn_pkt->rpn_val.bit_rate = rpn_pkt->fcs;
		return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
	} else {
		rpn_pkt->mcc_s_head.length.len = 8;
		memcpy(&(rpn_pkt->rpn_val), &rpn_val, sizeof(rpn_values));
		//print_data("",(u8*) rpn_pkt, rfcomm_frame_size);
		return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
	}
}

s32 
rfcomm_rls_msg(rfcomm_con *rfcomm, u8 cr, u8 dlci, u8 err_code)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	rls_msg* rls_pkt;
	u32 rfcomm_frame_size;
	u32 rfcomm_payload_size;

	rfcomm_frame_size = sizeof(rls_msg);
	rfcomm_payload_size = rfcomm_frame_size - sizeof(short_frame)-FCS_SIZE;
	rls_pkt = (rls_msg*) &tx_buf;
  
	set_uih_hdr((short_frame*) rls_pkt, CTRL_CHAN, rfcomm_payload_size,
		    rfcomm->initiator);
	rls_pkt->fcs = crc_calc((u8*) rls_pkt, SHORT_CRC_CHECK);
  
	rls_pkt->mcc_s_head.type.ea = EA;
	rls_pkt->mcc_s_head.type.cr = cr;
	rls_pkt->mcc_s_head.type.type = RLS;
	rls_pkt->mcc_s_head.length.ea = EA;
	rls_pkt->mcc_s_head.length.len = 2;
  
	rls_pkt->dlci.ea = EA;
	rls_pkt->dlci.cr = 1; /* Fixed value */
	rls_pkt->dlci.d = dlci & 1;
	rls_pkt->dlci.server_chn = dlci >> 1;
	rls_pkt->error = err_code;
	rls_pkt->res = 0;
  
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}
/* Sends an PN-messages and sets the not negotiable parameters to their
   default values in RFCOMM */

s32 
send_pn_msg(rfcomm_con *rfcomm, u8 prior, u32 frame_size, u8 credit_flow, u8 credits, u8 dlci, u8 cr)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	pn_msg *pn_pkt;
	u32 rfcomm_frame_size;
	D_CTRL("send_pn_msg: DLCI 0x%02x, prior:0x%02x, "
		"frame_size:%d, credit_flow:%x, credits:%d, cr:%x\n",
		dlci, prior, frame_size, credit_flow, credits, cr);

	rfcomm_frame_size = sizeof *pn_pkt;
	pn_pkt = (pn_msg*) &tx_buf;
	/* Set the UIH headers */
	set_uih_hdr((void*) pn_pkt, CTRL_CHAN, rfcomm_frame_size - 
		    (sizeof(short_frame) + FCS_SIZE), rfcomm->initiator);
	pn_pkt->fcs = crc_calc((u8*) pn_pkt, SHORT_CRC_CHECK);

	/* set the MCC-packet header */
	pn_pkt->mcc_s_head.type.ea = 1;
	pn_pkt->mcc_s_head.type.cr = cr;
	pn_pkt->mcc_s_head.type.type = PN;
	pn_pkt->mcc_s_head.length.ea = 1;
	/* The PN packet has a fix length of 8 bytes */
	pn_pkt->mcc_s_head.length.len = 8;

	/* Set the parameters in the PN-packet */
	pn_pkt->res1 = 0;
	pn_pkt->res2 = 0;
	pn_pkt->dlci = dlci;
	pn_pkt->frame_type = 0;
	pn_pkt->credit_flow = credit_flow;
	pn_pkt->prior = prior;
	pn_pkt->ack_timer = 0;
	pn_pkt->frame_size = cpu_to_le16(frame_size);
	pn_pkt->credits = credits;
	pn_pkt->max_nbrof_retrans = 0;
  
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

/* Sendares a Not supported command - command, which needs 3 bytes */

s32 
send_nsc_msg(rfcomm_con *rfcomm, mcc_type cmd, u8 cr) 
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	nsc_msg *nsc_pkt;
	u32 rfcomm_frame_size;
  
	rfcomm_frame_size = sizeof(nsc_msg);
	nsc_pkt = (nsc_msg*) &tx_buf;
 
	set_uih_hdr((void*) nsc_pkt, CTRL_CHAN, sizeof(nsc_msg)
		    - sizeof(short_frame) - FCS_SIZE, rfcomm->initiator);

	nsc_pkt->fcs = crc_calc((u8*) nsc_pkt, SHORT_CRC_CHECK);

	nsc_pkt->mcc_s_head.type.ea = 1;
	nsc_pkt->mcc_s_head.type.cr = cr;
	nsc_pkt->mcc_s_head.type.type = NSC;
	nsc_pkt->mcc_s_head.length.ea = 1;
	nsc_pkt->mcc_s_head.length.len = 1;

	nsc_pkt->command_type.ea = 1;
	nsc_pkt->command_type.cr = cmd.cr;
	nsc_pkt->command_type.type = cmd.type;

	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

s32
rfcomm_msc_msg(rfcomm_con *rfcomm, u8 value, u8 cr, u8 dlci)
{
	u8 tx_buf[MAX_TX_BUF_SIZE];
	msc_msg *msc_pkt;
	u32 rfcomm_frame_size;
  
	D_CTRL(FNC"val:%d, cr:%d, dlci:%d\n", value, cr, dlci);

	rfcomm_frame_size = sizeof(msc_msg);
	msc_pkt = (msc_msg*) &tx_buf;
 
	set_uih_hdr((void*) msc_pkt, CTRL_CHAN, sizeof(msc_msg)
		    - sizeof(short_frame) - FCS_SIZE, rfcomm->initiator);

	msc_pkt->fcs = crc_calc((u8*) msc_pkt, SHORT_CRC_CHECK);

	msc_pkt->mcc_s_head.type.ea = 1;
	msc_pkt->mcc_s_head.type.cr = cr;
	msc_pkt->mcc_s_head.type.type = MSC;
	msc_pkt->mcc_s_head.length.ea = 1;
	msc_pkt->mcc_s_head.length.len = 2;

	msc_pkt->dlci.ea = 1;
	msc_pkt->dlci.cr = 1;
	msc_pkt->dlci.d = dlci & 1;
	msc_pkt->dlci.server_chn = (dlci >> 1) & 0x1f;

	msc_pkt->v24_sigs = *(v24_signals *)&value;
	return write(rfcomm->iff->sock, tx_buf, rfcomm_frame_size);
}

void
set_uih_hdr(short_frame *uih_pkt, u8 dlci, u32 len, u8 cr)
{
	uih_pkt->h.addr.ea = 1;
	uih_pkt->h.addr.cr = cr;
	uih_pkt->h.addr.d = dlci & 0x1;
	uih_pkt->h.addr.server_chn = dlci >> 1;
	uih_pkt->h.control = CLR_PF(UIH);
	
	if (len > SHORT_PAYLOAD_SIZE) {
		((long_frame*) uih_pkt)->h.length.bits.ea = 0;
		((long_frame*) uih_pkt)->h.length.bits.len = len;  
	} else {
		uih_pkt->h.length.ea = 1;
		uih_pkt->h.length.len = len;  
	}

}


rfcomm_con* 
get_new_rfcomm_con(void)
{
	s32 i = 0;

	D_CTRL(FNC"rfcomm_con -> ttyBT%d\n",i);

        for (i = 0 ; i < BT_NBR_DATAPORTS ; i ++)
        {
           if (rfcomm_con_list[i].dlci[0].state == DISCONNECTED)
           {
              rfcomm_con_list[i].iff = NULL;
              return &rfcomm_con_list[i];
           }
        }
        return NULL;
}

rfcomm_con* 
get_rfcomm_con(u8 line)
{
	if(line >= BT_NBR_DATAPORTS) {
		return NULL;
	}
	return &rfcomm_con_list[line];
}

s32
valid_dlci(u8 dlci)
{
	if ((dlci < 62) && (dlci > 1)) {
		return TRUE;
	} else {
		return FALSE;
	}
}

/* fetches the 'last' non DISCONNECTED dlci number in this session, 
   returns -1 if no connected dlci was found */

s32 get_connected_dlci(rfcomm_con *rfcomm)
{
	s32 tmp;

	tmp = 61;
	while ((tmp >= 0) && (rfcomm != NULL) && 
	       (rfcomm->dlci[tmp].state == DISCONNECTED)) {
		tmp--;
	}
	return tmp;
}

/* Functions for the crc-check and calculation */

#define CRC_VALID 0xcf

/* This functions check whether the checksum is correct or not. Length is
   the number of bytes in the message, data points to the beginning of the
   message */

u32 
crc_check(u8 *data, u32 length, u8 check_sum)
{
	u8 fcs = 0xff;

	while (length--) {
		fcs = crctable[fcs ^ *data++];
	}
	fcs = crctable[fcs ^ check_sum];
	D_REC(FNC"fcs : %d\n", fcs);
	if (fcs == (uint) 0xcf) /* CRC_VALID) */ {
		D_REC(FNC"CRC check OK\n");
		return 0;
	} else {
		D_WARN(FNC"CRC check failed, data will be discarded\n");
		return -1;
	}
}

/* Calculates the checksum according to the RFCOMM specification */

u8 
crc_calc(u8 *data, u32 length)
{
	u8 fcs = 0xff;
  
	while (length--) {
		fcs = crctable[fcs^*data++];
	}
  
	/*Ones complement*/
	return 0xff-fcs;
}

/* Calculates a reversed CRC table for the FCS check */

void
create_crctable(u8 table[])
{
	s32 i,j;
  
	u8 data;
	u8 code_word = (u8) 0xe0;  /* pol = x8+x2+x1+1 */
	u8 sr = (u8) 0;            /* Shiftregister initiated to zero */

	for (j = 0; j < 256; j++) {
		data = (u8) j;
    
		for (i = 0; i < 8; i++) {
			if ((data & 0x1)^(sr & 0x1)) {
				sr >>= 1;
				sr ^= code_word;
			} else {
				sr >>= 1;
			}
      
			data >>= 1;
			sr &= 0xff;
		}
    
		table[j] = sr;
		sr = 0;
	} 
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/* /proc file functions */

#if 0

s32 rfcomm_sprint_con(u8 *buf)
{
	s32 i, j;
	s32 len=0;

	for (i = 0; i < BT_NBR_DATAPORTS; i++) {

		/* FIXME - move mtu etc into each dlci port, 
		   should not be in 'line' struct */
		
		for (j = 0; j < 62; j++){	
			if (rfcomm_con_list[i].dlci[j].state != DISCONNECTED){
				
				len+=sprintf(buf+len, "line[%d] mtu[%d] ",
					     rfcomm_con_list[i].line, 
					     rfcomm_con_list[i].dlci[j].mtu);
				
				if(j) {
					len +=sprintf(buf+len, " lc[%d] rc[%d] ",
						      rfcomm_con_list[i].dlci[j].local_credits,
						      rfcomm_con_list[i].dlci[j].remote_credits);
				}
				

				len+=sprintf(buf+len, "dlci#%d state[%s]\n", 
					     j, state2name(rfcomm_con_list[i].dlci[j].state));
			}
		}
	}
	return len;
}


s32 rfcomm_sprint_status(u8 *buf)
{
	s32 len=0;
	len+=sprintf(buf+len, "\n[RFCOMM]\n\n");
	len+=rfcomm_sprint_con(buf+len);
	return len;
}

#endif

s32
rfcomm_sabm_msg(rfcomm_con *rfcomm, u8 dlci)
{
	return send_sabm(rfcomm, dlci);
}

s32
rfcomm_pn_msg(rfcomm_con *rfcomm, u8 dlci, u8 credits, u32 frame_size)
{
	if (credits) {
		DSYS(FNC"using %d credits\n", credits);
		return send_pn_msg(rfcomm, 0, frame_size, 0xf, credits, dlci,
				   TRUE);
	} else {
		return send_pn_msg(rfcomm, 0, frame_size, 0, 0, dlci, TRUE);
	}
	
}

s32
rfcomm_ua_msg(rfcomm_con *rfcomm, u8 dlci)
{
	return send_ua(rfcomm, dlci);
}

s32
rfcomm_dm_msg(rfcomm_con *rfcomm, u8 dlci)
{
	return send_dm(rfcomm, dlci);
}

s32
rfcomm_disc_msg(rfcomm_con *rfcomm, u8 dlci)
{
	return send_disc(rfcomm, dlci);
}

/* Frame handler */
static volatile sig_atomic_t terminate = 0;

void rfcomm_sig_term(int sig)
{
	io_cancel();
	terminate = 1;
}

int rfcomm_frame_handler(rfcomm_con *conn)
{
	rfcomm_iff *iff = conn->iff;
	struct sigaction sa, osa;
	register int len;
	char buf[2048];
	fd_set fdset;
	int maxfd, idle = 0;
 
	io_init();

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = rfcomm_sig_term;
	sigaction(SIGTERM, &sa, &osa);

	maxfd = (iff->sock > iff->pty ? iff->sock : iff->pty) + 1;

	while( !terminate ){
		errno = 0;

		FD_ZERO(&fdset);
		FD_SET(iff->sock, &fdset);

		/* Don't monitor PTY unless we are connected */
		if( conn->dlci )
			FD_SET(iff->pty, &fdset);

		/* Wait for data */
		if( (len = select(maxfd, &fdset, NULL, NULL, NULL)) < 0 ) {
			if(errno != EAGAIN && errno != EINTR) {
				syslog(LOG_ERR, "select() failed: %s (%d)",
						strerror(errno), errno);
				break;
			} else
				continue;
		}

		if( FD_ISSET(iff->sock, &fdset) ){
			idle = 0;
			if( (len = read(iff->sock, buf, sizeof(buf))) < 0) {
				if(errno != EAGAIN && errno != EINTR) {
					syslog(LOG_ERR, 
						"read() failed: %s (%d)",
						strerror(errno), errno);
					break;
				} else
					continue;
			}
			if( !len ) {
				syslog(LOG_INFO, "End of stream detected");
				break;
			}

			rfcomm_receive_data(conn, buf, len);
		}

		if (FD_ISSET(iff->pty, &fdset)) {
			if ((len = pty_read(iff->pty, buf, sizeof(buf))) < 0 ) {
				if( errno != EAGAIN && errno != EINTR ) {
					syslog(LOG_ERR, 
						"pty_read() failed: %s (%d)",
						strerror(errno), errno);
					break;
				} else
					continue;
			}
			if( !len )
				continue;
			if( rfcomm_send_data(conn, buf, len) < 0 ) {
				syslog(LOG_ERR, "Send data failed %s(%d)",
						strerror(errno), errno);
				break;
			}
		}
	}

	send_disc(conn, conn->c_dlci);

	sigaction(SIGTERM, &osa, NULL);

	close(conn->iff->sock);
	close(conn->iff->pty);

	return terminate;
}

/* Open RFCOMM session */
int rfcomm_open(int line, int channel)
{
	struct l2cap_options opts;
	rfcomm_con *rfcomm;
	int i, opt;

	if( !(rfcomm = get_rfcomm_con(line)) || !rfcomm->iff )
		return -1;

	/* Get L2CAP options */
	opt = sizeof(opts);
	if( getsockopt(rfcomm->iff->sock, SOL_L2CAP, L2CAP_OPTIONS, &opts, &opt) < 0 ) {
		perror("Can't get L2CAP options.");
		return -1;
	}

	rfcomm->line 		= line;
	rfcomm->server_chn 	= channel;
	rfcomm->local_mtu	= opts.imtu;
	rfcomm->remote_mtu	= opts.omtu;

	/* Set MTU on all possible DLCIs */
	for(i = 0; i < 62; i++)
		rfcomm->dlci[i].mtu = (opts.omtu - 5);

	if( channel ){
		__u16 dlci;
		rfcomm->dlci[0].state = CONNECTING;
		rfcomm->initiator = TRUE;

		dlci = (channel << 1) | (~rfcomm->initiator & 0x1);
		rfcomm->dlci[dlci].initiator = TRUE;

		send_sabm(rfcomm, 0);
	}
	return rfcomm_frame_handler(rfcomm);
}

/* Close RFCOMM session */
void rfcomm_close(int line)
{
	rfcomm_con *rfcomm;
	int tmp;
	D_CTRL("rfcomm_disconnect_req %d\n", line);
	if( !(rfcomm = &rfcomm_con_list[line]) ){
		D_ERR("rfcomm_disconnect_req: no rfcomm con\n");
		return;
	}

	tmp = get_connected_dlci(rfcomm);

	if( tmp > 0 ){
		rfcomm->dlci[tmp].state = DISCONNECTING;
		send_disc(rfcomm, tmp);
	} else
		D_WARN("rfcomm_disconnect_req : line not connected !\n");
}

/* Register RFCOMM interface */
int rfcomm_register_iface(rfcomm_iff *iff)
{
	rfcomm_con *conn;

	if( !iff || !iff->sock || iff->sock < 0 )
		return -1;

	if( !(conn = get_new_rfcomm_con()) )
		return -1;

	conn->iff = iff;
	return conn->line;
}
/****************** END OF FILE rfcomm.c ************************************/
